; Pacman Assembly Script
;
; This is "proof of concept" for the VM - the HAPI only provides the drawing
; functions, input and frame limiting, everything else (logic, variables) is
; done in this script.
;
; You wouldn't actually use the VM in this way - it's intended for small
; blocks of simple code, with most of the work done by HAPI calls. Anyone who
; would actually write a script for the VM in the way I am here is completely
; insane.
;
; HAPI functions (remember to push parameters backwards!):
;	HapiDrawCell (x, y,  cellState)
;	HapiCreateSprite(), returns index
;	HapiPositionSprite(index, x, y)
;	HapiSetSpriteFrame(index, frame)
;	HapiFreeSprite(index)
;	HapiGetDelta(), returns delta as real
;	HapiRnd(), returns random number between 0 and 1
;	HapiGetTicks(), returns ticks as integer
;	HapiKeyDown(key), returns 0 (false) or 1 (true)
;	HapiDisplayStats(score, eaten)

Var	grid[1200] ; 40x30, 0 = blank, 1 = dot, 2 = wall, 3= bonus
Var	grid_count

Var	player_speed
Var	player_sprite
Var	player_px
Var	player_py
Var	player_dir
Var	player_targetx
Var	player_targety
Var	player_score
Var	player_dead
Var	player_eaten

Var	kup
Var	kdown
Var	kright
Var	kleft

Var	ghosts ; Increments after you win
Var	ghost_speed
Var	ghost_sprite[128]
Var	ghost_px[128]
Var	ghost_py[128]
Var	ghost_dir[128] ; Up = 0, Right = 1, Down = 2, Left = 3
Var	ghost_targetx[128]
Var	ghost_targety[128]
Var	ghost_state[128]

Func	Main
{
	Call	Reset
	Call	Init
	
	Mov	kup, 200
	Mov	kdown, 208
	Mov	kright, 205
	Mov	kleft, 203
}

Func	Reset
{
	Mov	ghosts, 8
	Mov	ghost_speed, 0.1
	Mov	player_speed, 0.15
	Mov	player_score, 0
	Mov	player_dead, 0
	Mov	player_eaten, 0
}

Func	Next
{
	Call	Free
	Call	Init
}

Func	Die
{
	Call	Free
	Call	Reset
	Call	Init
}

Func	Init
{

	Var	x
	Var	y
	Var	y2
	Var	xy
	
	Mov	x, 0
	Mov	y, 0
	Mov	xy, 0

	Mov	player_dead, 0
	Mov	player_eaten, 0

	; Border
	
	forloop10:	; Init the grid with dots
		Mov	grid[xy], 1
		Add	xy, 1
		Cmp	xy, 1200
		JL	forloop10
	
	Mov	x, 0
		
	forloop21:	; Horizontal Border 1
		Mov	grid[x], 2
		Add	x, 1
		Cmp	x, 40
		JL	forloop21

	Mov	x, 1120 ; 1200-40-40

	forloop22:	; Horizontal Border 2
		Mov	grid[x], 2
		Add	x, 1
		Cmp	x, 1160
		JL	forloop22
	
	Mov	y, 0
	
	forloop31:	; Vertical Border 1
		Mov	grid[y], 2
		Mov	y2, y
		Inc	y2
		;Mov	grid[y2], 2
		
		Add	y, 40
		Cmp	y, 1200
		JL	forloop31

	Mov	y, 39 ; 40-1

	forloop32:	; Vertical Border 2
		Mov	grid[y], 2
		Mov	y2, y
		Dec	y2
		;Mov	grid[y2], 2
		
		Add	y, 40
		Cmp	y, 1200
		JL	forloop32

	Mov	x, 1160 ; 1200-40-40

	forloop23:	; Clear bottom
		Mov	grid[x], 0
		Add	x, 1
		Cmp	x, 1200
		JL	forloop23
	
	Mov	xy, 0	
	
	; Extra rows
	
	Mov	y, 2
	
	; Left
	forloop101a:
		Mov	x, 2
	
		forloop101b:
			Push	x
			Push	y
			Call	toXy
			Mov	grid[_ret], 2
			Add	x, 1
			Cmp	x, 14
			JL 	forloop101b
		
		Add	y, 2
		Cmp	y, 28
		JL 	forloop101a

	Mov	y, 2	
	
	; middle
	forloop102a:
		Mov	x, 15
	
		forloop102b:
			Push	x
			Push	y
			Call	toXy
			Mov	grid[_ret], 2
			Add	x, 1
			Cmp	x, 24
			JL 	forloop102b
		
		Add	y, 2
		Cmp	y, 28
		JL 	forloop102a
	
	Mov	y, 14	
	
	; Bottom far-right
	forloop103a:
		Mov	x, 25
	
		forloop103b:
			Push	x
			Push	y
			Call	toXy
			Mov	grid[_ret], 2
			Add	x, 1
			Cmp	x, 38
			JL 	forloop103b
		
		Add	y, 2
		Cmp	y, 28
		JL 	forloop103a
	
	; Ghost Box - top & bottom
	Mov	x, 28
	
	forloop110:
		Push	x
		Push	5
		Call	toXy
		Mov	grid[_ret], 2
		
		Push	x
		Push	9
		Call	toXy
		Mov	grid[_ret], 2
		
		Add	x, 1
		Cmp	x, 35
		JL	forloop110
	
	; gap
	Push	31
	Push	5
	Call	toXy
	Mov	grid[_ret], 0

	Push	31
	Push	9
	Call	toXy
	Mov	grid[_ret], 0
		
	; Clear the dots in the ghost box
	
	Mov 	y, 6
	forloop112a:
		Mov	x, 28
	
		forloop112b:
			Push	x
			Push	y
			Call	toXy
			
			Push	x
			Push	y
			Call	toXy
			Mov	grid[_ret], 0
			Add	x, 1
			Cmp	x, 35
			JL 	forloop112b
		
		Add	y, 1
		Cmp	y, 9
		JL 	forloop112a

	Mov	y, 14
	
	; Ghost Box - individual

	Push	28
	Push	6
	Call	toXy
	Mov	grid[_ret], 2
	
	Push	28
	Push	8
	Call	toXy
	Mov	grid[_ret], 2
	
	Push	34
	Push	6
	Call	toXy
	Mov	grid[_ret], 2
	
	Push	34
	Push	8
	Call	toXy
	Mov	grid[_ret], 2
	
	; Individual blocks
	
	Push	2
	Push	27
	Call	toXy
	Mov	grid[_ret], 2

	Push	37
	Push	27
	Call	toXy
	Mov	grid[_ret], 2
	
	; Bonus dots!
	
	Push	1
	Push	1
	Call	toXy
	Mov	grid[_ret], 3

	Push	38
	Push	27
	Call	toXy
	Mov	grid[_ret], 3

	Push	1
	Push	27
	Call	toXy
	Mov	grid[_ret], 3

	Push	38
	Push	1
	Call	toXy
	Mov	grid[_ret], 3	
	
	; Ghosts
	Mov	x, 0
	
	forloop300:	; Init Ghosts!
	
		Hapi	HapiCreateSprite
		Mov	ghost_sprite[x], _ret
		Mov	ghost_px[x], 30.0
		Mov	ghost_py[x], 7.0
		Mov	ghost_targetx[x], 31.0
		Mov	ghost_targety[x], 7.0
		Mov	ghost_dir[x], 1
		Mov	ghost_state[x], 0
		
		Push	3
		Push	ghost_sprite[x]
		Hapi	HapiSetSpriteFrame
		
		Push    ghost_py[x]
		Push	ghost_px[x]
		Push	ghost_sprite[x]
		Hapi	HapiPositionSprite
		
		Add	x, 1
		Cmp	x, ghosts
		JL	forLoop300
	
	
	; Player
	Hapi	HapiCreateSprite
	Mov 	player_sprite, _ret
	Mov	player_px, 3.0
	Mov	player_py, 27.0
	Mov	player_dir, -1
	Mov	player_targetx, player_px
	Mov	player_targety, player_py
	
	; Clear player's cell
	Mov	x, player_px
	Mov	y, player_py
	Int	x
	Int	y
	Push	x
	Push	y
	Call	toXy
	Mov	grid[_ret], 0
	
	Push	5
	Push	player_sprite
	Hapi	HapiSetSpriteFrame
	
	Push	player_py
	Push	player_px
	Push	player_sprite
	Hapi	HapiPositionSprite
	
	
	; Apply & count
	Mov	xy, 0
	Mov	grid_count, 0
	
	forloop200:	; Make the HAPI render!
		
		; Get x
		Push 	xy
		Call	xyToX
		Mov	x, _ret

		; Get y
		Push 	xy
		Call	xyToY
		Mov	y, _ret
		
		; Draw
		Push	grid[xy]
		Push 	y
		Push 	x
		HAPI	HapiDrawCell
		
		; Count
		Cmp	grid[xy], 1
		JNE	continueForLoop200
		Inc	grid_count
		
		continueForLoop200:
		
		Add	xy, 1
		Cmp	xy, 1200
		JL	forloop200
	
		
}

Func	Update ; Called once per main loop
{
	Var	i
	Var	rnd
	Var	delta
	Var	ticks
	Var	deltaGhostSpeed
	Var	deltaPlayerSpeed
	
	Var	tmpx
	Var	tmpy
	Var	tmpxy
	Var	tmpxy2
	
	Var	tmpFrame
	
	
	; Delta
	Hapi	HapiGetDelta
	Mov	delta, _ret
	
	Hapi	HapiGetTicks
	Mov	ticks, _ret
	
	Mov	deltaGhostSpeed, delta
	Mul	deltaGhostSpeed, ghost_speed
	
	Mov	deltaPlayerSpeed, delta
	Mul	deltaPlayerSpeed, player_speed
	
	; Update Player
	
	; Input:
	
	; Up
	Push	kup
	Hapi	HapiKeyDown
	Cmp	_ret, 1
	JE	playerSetDir0

	; Right
	Push	kright
	Hapi	HapiKeyDown
	Cmp	_ret, 1
	JE	playerSetDir1
	
	; Down
	Push	kdown
	Hapi	HapiKeyDown
	Cmp	_ret, 1
	JE	playerSetDir2
	
	; Left
	Push	kleft
	Hapi	HapiKeyDown
	Cmp	_ret, 1
	JE	playerSetDir3
	
	Jmp	playerSetDirEnd2


	playerSetDir0:
		Mov	player_dir, 0
		Mov	player_targetx, player_px
		Mov	player_targety, player_py
		Dec	player_targety
		Jmp	playerSetDirEnd1
	
	playerSetDir1:
		Mov	player_dir, 1
		Mov	player_targetx, player_px
		Mov	player_targety, player_py
		Inc	player_targetx
		Jmp	playerSetDirEnd1
	
	playerSetDir2:
		Mov	player_dir, 2
		Mov	player_targetx, player_px
		Mov	player_targety, player_py
		Inc	player_targety
		Jmp	playerSetDirEnd1
		
	playerSetDir3:
		Mov	player_dir, 3
		Mov	player_targetx, player_px
		Mov	player_targety, player_py
		Dec	player_targetx
		Jmp	playerSetDirEnd1
	
	playerSetDirEnd1:
	
		Int	player_px
		Int	player_py
		Float	player_px
		Float	player_py
	
	playerSetDirEnd2:
	
	
	; Move
	Cmp	player_dir, 0
	JE	casePlayerUp
	Cmp	player_dir, 1
	JE	casePlayerRight
	Cmp	player_dir, 2
	JE	casePlayerDown
	Cmp	player_dir, 3
	JE	casePlayerLeft
	Jmp	casePlayerStill
	
		casePlayerUp:
			Sub	player_py, deltaPlayerSpeed
			Cmp	player_py, player_targety
			JL	casePlayerUp2
			Jmp	casePlayerEnd1
				
				casePlayerUp2:
					Dec	player_targety
					Jmp	playerContinue
		
		casePlayerRight:
			Add	player_px, deltaPlayerSpeed
			Cmp	player_px, player_targetx
			JG	casePlayerRight2
			Jmp	casePlayerEnd1
				
				casePlayerRight2:
					Inc	player_targetx
					Jmp	playerContinue
		
		casePLayerDown:
			Add	player_py, deltaPlayerSpeed
			Cmp	player_py, player_targety
			JG	casePlayerDown2
			Jmp	casePlayerEnd1
				
				casePlayerDown2:
					Inc	player_targety
					Jmp	playerContinue
		
		casePlayerLeft:
			Sub	player_px, deltaPlayerSpeed
			Cmp	player_px, player_targetx
			JL	casePlayerLeft2
			Jmp	casePlayerEnd1
				
				casePlayerLeft2:
					Dec	player_targetx
					Jmp	playerContinue
		
		casePlayerStill:
			Jmp	playerContinue
						
	
	casePlayerEnd1:
	Jmp	playerContinue
	
	
	playerContinue:
	
	checkPlayerCollisions:
		; Get grid xy			
		Mov	tmpx, player_targetx
		Mov	tmpy, player_targety
		Int	tmpx
		Int	tmpy
		Push	tmpx
		Push	tmpy
		Call	toXy
		Mov	tmpxy, _ret
		
		; Check for wall
		Cmp	grid[tmpxy], 2
		JE	playerStop
		Jmp	playerContinue2
	
	playerStop:
		Int	player_px
		Int	player_py
		Float	player_px
		Float	player_py
		Mov	player_targetx, player_px
		Mov	player_targety, player_py
		Mov	player_dir, 4	
	
	
	playerContinue2:
	
	
	; Check for pills
	Mov	tmpx, player_px
	Mov	tmpy, player_py
	Int	tmpx
	Int	tmpy
	Push	tmpx
	Push	tmpy
	Call	toXy
	Cmp	grid[_ret], 1
	JE	playerEatPill
	Cmp	grid[_ret], 3
	JE	playerEatPillBonus
	Jmp	playerEatPillEnd
	
		playerEatPill:
			Inc	player_score
			Inc	player_eaten
			Mov	grid[_ret], 0
			Push	grid[_ret]
			Push 	tmpy
			Push 	tmpx
			HAPI	HapiDrawCell
			Jmp	playerEatPillEnd
			
		playerEatPillBonus:
			Mov	i, 0
			Add	player_score, 10
			Mov	grid[_ret], 0
			Push	grid[_ret]
			Push 	tmpy
			Push 	tmpx
			HAPI	HapiDrawCell
			
			forBonus1:
				Mov	ghost_state[i], 1
			
				Inc	i
				Cmp	i, ghosts
				JL	forBonus1
		
		playerEatPillEnd:
	
	; Check for ghosts
	Mov	i, 0
	forPlayerVsGhost1:

		Push	ghost_px[i]
		Push	ghost_py[i]
		Push	player_px
		Push	player_py
		Call	PlayerGhostCollision
		
		Cmp	_ret, 1
		JE	gpCollided
		Jmp	gpCollidedEnd
		
		gpCollided:
			Cmp	ghost_state[i], 1
			JE	eatGhost
			Mov	player_dead, 1			
			Jmp	gpCollidedEnd
			
			eatGhost:
				Add	player_score, 10
				Mov	ghost_state[i], 0
				Mov	ghost_px[i], 30.0
				Mov	ghost_py[i], 7.0
				
		
		gpCollidedEnd:
			
		Inc	i
		Cmp	i, ghosts
		JL	forPlayerVsGhost1
		
	
	; Update sprite position
	Push	player_py
	Push	player_px
	Push	player_sprite
	Hapi	HapiPositionSprite
	
	; Update sprite frame
	Mov	tmpFrame, 0
	Cmp	player_dir, 0
	JE	caseFrameUp
	Cmp	player_dir, 1
	JE	caseFrameRight
	Cmp	player_dir, 2
	JE	caseFrameDown
	Cmp	player_dir, 3
	JE	caseFrameLeft
	Jmp	caseFrameRight
	
	caseFrameUp:
		Mov	tmpFrame, 8
		Jmp	caseFrameEnd
		
	caseFrameRight:
		Mov	tmpFrame, 5
		Jmp	caseFrameEnd
	
	caseFrameDown:
		Mov	tmpFrame, 6
		Jmp	caseFrameEnd
	
	caseFrameLeft:
		Mov	tmpFrame, 7
		Jmp	caseFrameEnd
		
	
	caseFrameEnd:
	
	; Animation
	Mov	i, ticks
	Mod	i, 50
	Cmp	i, 25
	JL	frameAdd1
	Jmp	frameAdd0
	
	frameAdd1:
		Add	tmpFrame, 5
	frameAdd0:
	
	; Set Frame
	Push	tmpFrame
	Push	player_sprite
	Hapi	HapiSetSpriteFrame
	
	
	
	
	
	; Update Ghosts
	Mov	i, 0
	forGhosts1:

		; Move
		Cmp	ghost_dir[i], 0
		JE	caseGhostUp
		Cmp	ghost_dir[i], 1
		JE	caseGhostRight
		Cmp	ghost_dir[i], 2
		JE	caseGhostDown
		Jmp	caseGhostLeft
		
		caseGhostUp:
			Sub	ghost_py[i], deltaGhostSpeed
			Cmp	ghost_py[i], ghost_targety[i]
			JL	caseGhostUp2
			Jmp	caseGhostEnd1
			
			caseGhostUp2:
				Dec	ghost_targety[i]
				Jmp	checkGhostCollisions
		
		caseGhostRight:		
			Add	ghost_px[i], deltaGhostSpeed
			Cmp	ghost_px[i], ghost_targetx[i]
			JG	caseGhostRight2
			Jmp	caseGhostEnd1
			
			caseGhostRight2:
				Inc	ghost_targetx[i]
				Jmp	checkGhostCollisions
		
		caseGhostDown:
			Add	ghost_py[i], deltaGhostSpeed
			Cmp	ghost_py[i], ghost_targety[i]
			JG	caseGhostDown2
			Jmp	caseGhostEnd1
			
			caseGhostDown2:
				Inc	ghost_targety[i]
				Jmp	checkGhostCollisions
		
		caseGhostLeft:
			Sub	ghost_px[i], deltaGhostSpeed
			Cmp	ghost_px[i], ghost_targetx[i]
			JL	caseGhostLeft2
			Jmp	caseGhostEnd1
			
			caseGhostLeft2:
				Dec	ghost_targetx[i]
				Jmp	checkGhostCollisions
		
		; End Case
		caseGhostEnd1:
		Jmp	ghostContinue
		
		; Check Collisions
		checkGhostCollisions:
				
			; Get grid xy			
			Mov	tmpx, ghost_targetx[i]
			Mov	tmpy, ghost_targety[i]
			Int	tmpx
			Int	tmpy
			Push	tmpx
			Push	tmpy
			Call	toXy
			Mov	tmpxy, _ret
			
			; Randomly change direction
			Hapi	HapiRnd
			Mov	rnd, _ret
			Cmp	rnd, 0.25
			JLE	ghostNewDirection
			
			; Check for wall
			Cmp	grid[tmpxy], 2
			JE	ghostNewDirection
			Jmp	ghostContinue
		
		; New Direciton
		ghostNewDirection:
		
			; Reset targets
			Mov	ghost_targetx[i], ghost_px[i]
			Mov	ghost_targety[i], ghost_py[i]
			Int	ghost_targetx[i]
			Int	ghost_targety[i]
			
			; Get grid xy			
			Mov	tmpx, ghost_targetx[i]
			Mov	tmpy, ghost_targety[i]
			Int	tmpx
			Int	tmpy
			Push	tmpx
			Push	tmpy
			Call	toXy
			Mov	tmpxy, _ret
			
			; Choose a direction			
			ghostDirChoose:

				Mov	tmpxy2, tmpxy
				
				Hapi	HapiRnd
				Mov	rnd, _ret
				
				Cmp	rnd, 0.25
				JLE	ghostTryDirUp
				Cmp	rnd, 0.50
				JLE	ghostTryDirRight
				Cmp	rnd, 0.75
				JLE	ghostTryDirDown
				Jmp	ghostTryDirLeft
			
				ghostTryDirUp:		
					Sub	tmpxy2, 40					
					Cmp	grid[tmpxy2], 2
					JE	ghostDirChoose
					Mov	ghost_dir[i], 0
					Mov	tmpy, ghost_py[i]
					Int	tmpy
					Dec	tmpy
					Mov	ghost_targety[i], tmpy
					Jmp	checkGhostCollisions
									
				ghostTryDirRight:	
					Add	tmpxy2, 1					
					Cmp	grid[tmpxy2], 2
					JE	ghostDirChoose
					Mov	ghost_dir[i], 1
					Mov	tmpx, ghost_px[i]
					Int	tmpx
					Inc	tmpx
					Mov	ghost_targetx[i], tmpx
					Jmp	checkGhostCollisions
				
				ghostTryDirDown:
					Add	tmpxy2, 40
					Cmp	grid[tmpxy2], 2
					JE	ghostDirChoose
					Mov	ghost_dir[i], 2
					Mov	tmpy, ghost_py[i]
					Int	tmpy
					Inc	tmpy
					Mov	ghost_targety[i], tmpy
					Jmp	checkGhostCollisions
				
				ghostTryDirLeft:
					Sub	tmpxy2, 1
					Cmp	grid[tmpxy2], 2
					JE	ghostDirChoose
					Mov	ghost_dir[i], 3
					Mov	tmpx, ghost_px[i]
					Int	tmpx
					Dec	tmpx
					Mov	ghost_targetx[i], tmpx
					Jmp	checkGhostCollisions
		
		
		ghostContinue:
		
		; Update sprite Position
		Push	ghost_py[i]
		Push	ghost_px[i]
		Push	ghost_sprite[i]
		Hapi	HapiPositionSprite
		
		; Update sprite frame
		Cmp	ghost_state[i], 0
		JE	ghostFrame1
		Jmp	ghostFrame2
		
		ghostFrame1:
			Push	3
			Push	ghost_sprite[i]
			Hapi	HapiSetSpriteFrame
			Jmp	ghostFrameEnd
		
		ghostFrame2:
			Push	4
			Push	ghost_sprite[i]
			Hapi	HapiSetSpriteFrame
		
		ghostFrameEnd:
		
	
		Inc	i
		Cmp	i, ghosts
		JL	forGhosts1
	
	
	; Check for death
	Cmp	player_dead, 1
	JNE	notDead
		Call	Die
		Jmp	final
	
	notDead:
		Cmp	grid_count, player_eaten
		JNE	final
		Call	Next
		Jmp	final
	
	final:
	
	Push	player_eaten
	Push	player_score
	Hapi	HapiDisplayStats
	
	Ret	
}

Func	Free
{
	Var	i
	Mov	i, 0
	
	; Free Ghost Sprites
	freeLoop1:
	
		Push ghost_sprite[i]
		Hapi HapiFreeSprite
	
		Inc	i
		Cmp	i, ghosts
		JL	freeLoop1
		

	; Add more ghosts for next time
	Add	ghosts, 8
	Cmp	ghosts, 128
	JGE	ghostClamp
	Jmp	ghostClampEnd
	
	ghostClamp:
		Mov	ghosts, 128
		Jmp	ghostClampEnd
	
	ghostClampEnd:
	
	; Free Player Sprite
	Push	player_sprite
	Hapi	HapiFreeSprite
	
	; Speed it up!
	;Add	ghost_speed, 0.1
}

Func PlayerGhostCollision
{
	Param	py ; player
	Param	px
	Param	gy ; ghost
	Param	gx
	
	Int	py
	Int	px
	Int	gy
	Int	gx

	Mov	_ret, 0
	
	Cmp	py, gy
	JE	pgc1
	Ret
	
	pgc1:
	Cmp	px, gx
	JE	pgc2
	Ret
	
	pgc2:
	Mov	_ret, 1
}

Func	xyToX ; extracts x from xy value
{
	Param xy
	
	Mov _ret, xy
	Mod _ret, 40
}

Var xytoY_tmpX ; BUG - seems having both var and param in a function is buggy!
Func	xyToY ; extracts y from xy value
{	
	Param	xy	
	
	Mov	xytoY_tmpX, xy
	Mod	xytoY_tmpX, 40
	
	Mov	_ret, xy
	Sub	_ret, xytoY_tmpX
	Div	_ret, 40
}

Func	toXy ; Uses x,y to get xy value
{
	Param	y
	Param	x
	
	Mov	_ret, x
	Mul	y, 40
	Add	_ret, y
}
